@@ -196,6 +196,89 @@ int stream_setsockopt_nodelay(int fd, int proto, int on)
 	return rc;
 }
 
+int stream_setsockopt_tcp_keepalive(int fd, int on)
+{
+	int ret = setsockopt(fd, SOL_SOCKET, SO_KEEPALIVE, &on, sizeof(on));
+	if (ret < 0) {
+		ret = errno;
+		LOGP(DLINP, LOGL_ERROR, "Failed to enable TCP keepalive on fd %d: %s\n", fd, strerror(ret));
+		return -ret;
+	}
+	return 0;
+}
+
+int stream_setsockopt_tcp_keepidle(int fd, int keepalive_time)
+{
+	int ret = setsockopt(fd, IPPROTO_TCP, TCP_KEEPIDLE, &keepalive_time, sizeof(keepalive_time));
+	if (ret < 0) {
+		ret = errno;
+		LOGP(DLINP, LOGL_ERROR, "Failed to set TCP_KEEPIDLE on fd %d: %s\n", fd, strerror(ret));
+		return -ret;
+	}
+	return 0;
+}
+
+int stream_setsockopt_tcp_keepintvl(int fd, int keepalive_intvl)
+{
+	int ret = setsockopt(fd, IPPROTO_TCP, TCP_KEEPINTVL, &keepalive_intvl, sizeof(keepalive_intvl));
+	if (ret < 0) {
+		ret = errno;
+		LOGP(DLINP, LOGL_ERROR, "Failed to set TCP_KEEPINTVL on fd %d: %s\n", fd, strerror(ret));
+		return -ret;
+	}
+	return 0;
+}
+
+int stream_setsockopt_tcp_keepcnt(int fd, int keepalive_probes)
+{
+	int ret = setsockopt(fd, IPPROTO_TCP, TCP_KEEPCNT, &keepalive_probes, sizeof(keepalive_probes));
+	if (ret < 0) {
+		ret = errno;
+		LOGP(DLINP, LOGL_ERROR, "Failed to set TCP_KEEPCNT on fd %d: %s\n", fd, strerror(ret));
+		return -ret;
+	}
+	return 0;
+}
+
+int stream_setsockopt_tcp_user_timeout(int fd, unsigned int user_timeout)
+{
+	int ret = setsockopt(fd, IPPROTO_TCP, TCP_USER_TIMEOUT, &user_timeout, sizeof(user_timeout));
+	if (ret < 0) {
+		ret = errno;
+		LOGP(DLINP, LOGL_ERROR, "Failed to set TCP_USER_TIMEOUT on fd %d: %s\n", fd, strerror(ret));
+		return -ret;
+	}
+	return 0;
+}
+
+int stream_tcp_keepalive_pars_apply(int fd, const struct stream_tcp_keepalive_pars *tkp)
+{
+	int ret;
+
+	if (!tkp->enable)
+		return 0;
+
+	if ((ret = stream_setsockopt_tcp_keepalive(fd, tkp->enable)) < 0)
+		return ret;
+
+	if (tkp->time_present) {
+		if ((ret = stream_setsockopt_tcp_keepidle(fd, tkp->time_value)) < 0)
+			return ret;
+	}
+
+	if (tkp->intvl_present) {
+		if ((ret = stream_setsockopt_tcp_keepintvl(fd, tkp->intvl_value)) < 0)
+			return ret;
+	}
+
+	if (tkp->probes_present) {
+		if ((ret = stream_setsockopt_tcp_keepcnt(fd, tkp->probes_value)) < 0)
+			return ret;
+	}
+
+	return 0;
+}
+
 #ifdef HAVE_LIBSCTP
 static int stream_sctp_recvmsg_trailer(const char *log_pfx, struct msgb *msg, int ret, const struct sctp_sndrcvinfo *sinfo, int flags)
 {
@@ -266,8 +349,11 @@ static int stream_sctp_recvmsg_trailer(const char *log_pfx, struct msgb *msg, in
 		return ret;
 	}
 
-	if (OSMO_UNLIKELY(ret > 0 && !sinfo))
-		LOGP(DLINP, LOGL_ERROR, "%s sctp_recvmsg without SNDRCV cmsg?!?\n", log_pfx);
+	if (OSMO_UNLIKELY(ret > 0 && !sinfo)) {
+		char buf[512];
+		LOGP(DLINP, LOGL_ERROR, "%s sctp_recvmsg without SNDRCV cmsg?!? ret=%d, flags=0x%x, %s\n",
+		     log_pfx, ret, flags, msgb_hexdump_buf(buf, sizeof(buf), msg));
+	}
 
 	return ret;
 }
@@ -279,8 +365,14 @@ int stream_sctp_recvmsg_wrapper(int fd, struct msgb *msg, const char *log_pfx)
 	int flags = 0;
 	int ret;
 
+	/* Canary to detect if kernel returns sinfo; see https://github.com/sctp/lksctp-tools/issues/37 */
+	sinfo.sinfo_assoc_id = 0;
+
 	ret = sctp_recvmsg(fd, msg->tail, msgb_tailroom(msg), NULL, NULL, &sinfo, &flags);
-	return stream_sctp_recvmsg_trailer(log_pfx, msg, ret, &sinfo, flags);
+	if (sinfo.sinfo_assoc_id)
+		return stream_sctp_recvmsg_trailer(log_pfx, msg, ret, &sinfo, flags);
+	else
+		return stream_sctp_recvmsg_trailer(log_pfx, msg, ret, NULL, flags);
 }
 
 /*! wrapper for osmo_io asynchronous recvmsg response */
@@ -290,7 +382,7 @@ int stream_iofd_sctp_recvmsg_trailer(struct osmo_io_fd *iofd, struct msgb *msg,
 	struct cmsghdr *cmsg = NULL;
 
 	for (cmsg = CMSG_FIRSTHDR((struct msghdr *) msgh); cmsg != NULL;
-		cmsg = CMSG_NXTHDR((struct msghdr *) msgh, cmsg)) {
+	     cmsg = CMSG_NXTHDR((struct msghdr *) msgh, cmsg)) {
 		if (cmsg->cmsg_level == IPPROTO_SCTP && cmsg->cmsg_type == SCTP_SNDRCV) {
 			sinfo = (const struct sctp_sndrcvinfo *)CMSG_DATA(cmsg);
 			break;
