@@ -99,6 +99,7 @@ struct osmo_stream_srv_link {
 	unsigned int		tx_queue_max_length; /* Max amount of msgbs which can be enqueued */
 	struct msgb_alloc_info	msgb_alloc;
 	struct osmo_sock_init2_multiaddr_pars ma_pars;
+	struct stream_tcp_pars tcp_pars;
 };
 
 static int _setsockopt_nosigpipe(struct osmo_stream_srv_link *link, int new_fd)
@@ -142,11 +143,25 @@ static int osmo_stream_srv_link_ofd_cb(struct osmo_fd *ofd, unsigned int what)
 		LOGSLNK(link, LOGL_INFO, "accept()ed new link from %s\n",
 			osmo_sockaddr_to_str(&osa));
 
-		if (link->proto == IPPROTO_SCTP) {
+		switch (link->proto) {
+		case IPPROTO_TCP:
+			ret = stream_tcp_keepalive_pars_apply(sock_fd, &link->tcp_pars.ka);
+			if (ret < 0) {
+				LOGSLNK(link, LOGL_ERROR, "failed applying TCP keep-alive pars on fd %d\n", sock_fd);
+				goto error_close_socket;
+			}
+			/* tcp_pars.user_timeout (sockopt TCP_USER_TIMEOUT) is
+			 * inherited by accept() connected sockets automatically,
+			 * no need to re-apply it here. */
+			break;
+		case IPPROTO_SCTP:
 			_setsockopt_nosigpipe(link, sock_fd);
 			ret = stream_sctp_sock_activate_events(sock_fd);
 			if (ret < 0)
 				goto error_close_socket;
+			break;
+		default:
+			break;
 		}
 		break;
 	default:
@@ -628,7 +643,7 @@ void osmo_stream_srv_link_close(struct osmo_stream_srv_link *link)
 }
 
 /*! Set given parameter of stream_srv_link to given value.
- *  \param[in] cli stream client on which to set parameter.
+ *  \param[in] link stream server link on which to set parameter.
  *  \param[in] par identifier of the parameter to be set.
  *  \param[in] val value of the parameter to be set.
  *  \param[in] val_len length of the parameter value.
@@ -670,6 +685,43 @@ int osmo_stream_srv_link_set_param(struct osmo_stream_srv_link *link, enum osmo_
 		link->ma_pars.sctp.sockopt_initmsg.max_instreams_present = true;
 		link->ma_pars.sctp.sockopt_initmsg.max_instreams_value = *(uint16_t *)val;
 		break;
+	/* TCP keepalive params: */
+	case OSMO_STREAM_SRV_LINK_PAR_TCP_SOCKOPT_KEEPALIVE:
+		if (!val || val_len != sizeof(uint8_t))
+			return -EINVAL;
+		link->tcp_pars.ka.enable = !!*(uint8_t *)val;
+		/* Will be applied on accepted sockets */
+		break;
+	case OSMO_STREAM_SRV_LINK_PAR_TCP_SOCKOPT_KEEPIDLE:
+		if (!val || val_len != sizeof(int))
+			return -EINVAL;
+		link->tcp_pars.ka.time_present = true;
+		link->tcp_pars.ka.time_value = *(int *)val;
+		/* Will be applied on accepted sockets */
+		break;
+	case OSMO_STREAM_SRV_LINK_PAR_TCP_SOCKOPT_KEEPINTVL:
+		if (!val || val_len != sizeof(int))
+			return -EINVAL;
+		link->tcp_pars.ka.intvl_present = true;
+		link->tcp_pars.ka.intvl_value = *(int *)val;
+		/* Will be applied on accepted sockets */
+		break;
+	case OSMO_STREAM_SRV_LINK_PAR_TCP_SOCKOPT_KEEPCNT:
+		if (!val || val_len != sizeof(int))
+			return -EINVAL;
+		link->tcp_pars.ka.probes_present = true;
+		link->tcp_pars.ka.probes_value = *(int *)val;
+		/* Will be applied on accepted sockets */
+		break;
+	case OSMO_STREAM_SRV_LINK_PAR_TCP_SOCKOPT_USER_TIMEOUT:
+		if (!val || val_len != sizeof(unsigned int))
+			return -EINVAL;
+		link->tcp_pars.user_timeout_present = true;
+		link->tcp_pars.user_timeout_value = *(int *)val;
+		/* This value is inherited by accept() connected sockets (hence by child stream_srv): */
+		if (osmo_stream_srv_link_is_opened(link))
+			return stream_setsockopt_tcp_user_timeout(osmo_stream_srv_link_get_fd(link),
+								  link->tcp_pars.user_timeout_value);
 	default:
 		return -ENOENT;
 	};
@@ -691,6 +743,7 @@ struct osmo_stream_srv {
 	};
 	struct llist_head		tx_queue; /* osmo_ofd mode (only): Queue of msgbs */
 	unsigned int			tx_queue_count; /* osmo_ofd mode (only): Current amount of msgbs queued */
+	unsigned int			tx_queue_max_length; /* Max amount of msgbs which can be enqueued */
 	osmo_stream_srv_closed_cb_t	closed_cb;
 	osmo_stream_srv_read_cb_t	read_cb;
 	osmo_stream_srv_read_cb2_t	iofd_read_cb;
@@ -916,7 +969,10 @@ static int osmo_stream_srv_cb(struct osmo_fd *ofd, unsigned int what)
  *  \param[in] read_cb Call-back to call when the socket is readable
  *  \param[in] closed_cb Call-back to call when the connection is closed
  *  \param[in] data User data to save in the new Stream Server struct
- *  \returns Stream Server in case of success; NULL on error */
+ *  \returns Stream Server in case of success; NULL on error
+ *
+ *  Note: In case of error, the caller is responsible for closing the fd.
+ */
 struct osmo_stream_srv *
 osmo_stream_srv_create(void *ctx, struct osmo_stream_srv_link *link, int fd,
 		       osmo_stream_srv_read_cb_t read_cb,
@@ -941,6 +997,8 @@ osmo_stream_srv_create(void *ctx, struct osmo_stream_srv_link *link, int fd,
 
 	osmo_sock_get_name_buf(conn->sockname, sizeof(conn->sockname), fd);
 
+	conn->tx_queue_max_length = conn->srv->tx_queue_max_length;
+
 	if (osmo_fd_register(&conn->ofd) < 0) {
 		LOGSSRV(conn, LOGL_ERROR, "could not register FD\n");
 		talloc_free(conn);
@@ -959,7 +1017,10 @@ osmo_stream_srv_create(void *ctx, struct osmo_stream_srv_link *link, int fd,
  *  \param[in] link Stream Server Link to which we belong
  *  \param[in] fd system file descriptor of the new connection
  *  \param[in] data User data to save in the new Stream Server struct
- *  \returns Stream Server in case of success; NULL on error */
+ *  \returns Stream Server in case of success; NULL on error
+ *
+ *  Note: In case of error, the caller is responsible for closing the fd.
+ */
 struct osmo_stream_srv *
 osmo_stream_srv_create2(void *ctx, struct osmo_stream_srv_link *link, int fd, void *data)
 {
@@ -991,7 +1052,8 @@ osmo_stream_srv_create2(void *ctx, struct osmo_stream_srv_link *link, int fd, vo
 		return NULL;
 	}
 
-	osmo_iofd_set_txqueue_max_length(conn->iofd, conn->srv->tx_queue_max_length);
+	conn->tx_queue_max_length = conn->srv->tx_queue_max_length;
+	osmo_iofd_set_txqueue_max_length(conn->iofd, conn->tx_queue_max_length);
 	if (conn->srv->msgb_alloc.set_by_user)
 		osmo_iofd_set_alloc_info(conn->iofd, conn->srv->msgb_alloc.size, conn->srv->msgb_alloc.headroom);
 
@@ -1011,9 +1073,31 @@ osmo_stream_srv_create2(void *ctx, struct osmo_stream_srv_link *link, int fd, vo
  */
 void osmo_stream_srv_set_name(struct osmo_stream_srv *conn, const char *name)
 {
-	osmo_talloc_replace_string(conn, &conn->name, name);
+	osmo_stream_srv_set_name_f(conn, "%s", name);
+}
+
+/*! Set a name on the srv object using arguments like printf() (used during logging).
+ *  \param[in] srv stream_srv whose name is to be set
+ *  \param[in] name the name to be set on srv
+ */
+void osmo_stream_srv_set_name_f(struct osmo_stream_srv *conn, const char *fmt, ...)
+{
+	char *name = NULL;
+
+	if (fmt) {
+		va_list ap;
+
+		va_start(ap, fmt);
+		name = talloc_vasprintf(conn, fmt, ap);
+		va_end(ap);
+	}
+
+	if (conn->name)
+		talloc_free((void *)conn->name);
+	conn->name = name;
+
 	if (conn->mode == OSMO_STREAM_MODE_OSMO_IO && conn->iofd)
-		osmo_iofd_set_name(conn->iofd, name);
+		osmo_iofd_set_name(conn->iofd, conn->name);
 }
 
 /*! Retrieve name previously set on the srv object (see osmo_stream_srv_set_name()).
@@ -1153,9 +1237,24 @@ void *osmo_stream_srv_get_data(struct osmo_stream_srv *conn)
 	return conn->data;
 }
 
+/*! Set the maximum length queue of the stream server connection.
+ *  \param[in] conn Stream Server to modify
+ *  \param[in] size maximum amount of msgbs which can be queued in the internal tx queue.
+ *  \returns 0 on success, negative on error.
+ *
+ *  The default queue size of a osmo_stream_srv is inherited during creation time from
+ *  osmo_stream_srv_link. */
+int osmo_stream_srv_set_tx_queue_max_length(struct osmo_stream_srv *conn, unsigned int size)
+{
+	conn->tx_queue_max_length = size;
+	if (conn->mode == OSMO_STREAM_MODE_OSMO_IO && conn->iofd)
+		osmo_iofd_set_txqueue_max_length(conn->iofd, conn->tx_queue_max_length);
+	return 0;
+}
+
 /*! Retrieve the stream server socket description.
  *  The returned name is stored in a static buffer; it is hence not re-entrant or thread-safe!
- *  \param[in] cli Stream Server to examine
+ *  \param[in] conn Stream Server to examine
  *  \returns Socket description or NULL in case of error */
 const char *osmo_stream_srv_get_sockname(const struct osmo_stream_srv *conn)
 {
@@ -1261,7 +1360,7 @@ void osmo_stream_srv_send(struct osmo_stream_srv *conn, struct msgb *msg)
 
 	switch (conn->mode) {
 	case OSMO_STREAM_MODE_OSMO_FD:
-		if (conn->tx_queue_count >= conn->srv->tx_queue_max_length) {
+		if (conn->tx_queue_count >= conn->tx_queue_max_length) {
 			LOGSSRV(conn, LOGL_ERROR, "send: tx queue full, dropping msg!\n");
 			msgb_free(msg);
 			return;
@@ -1372,4 +1471,51 @@ void osmo_stream_srv_clear_tx_queue(struct osmo_stream_srv *conn)
 		osmo_stream_srv_destroy(conn);
 }
 
+/*! Set given parameter of stream_srv to given value.
+ *  \param[in] conn stream server conn on which to set parameter.
+ *  \param[in] par identifier of the parameter to be set.
+ *  \param[in] val value of the parameter to be set.
+ *  \param[in] val_len length of the parameter value.
+ *  \returns 0 in success; negative -errno on error. */
+int osmo_stream_srv_set_param(struct osmo_stream_srv *conn, enum osmo_stream_srv_param par,
+			      void *val, size_t val_len)
+{
+	uint8_t on;
+	int i;
+	unsigned int u;
+	OSMO_ASSERT(conn);
+
+	switch (par) {
+	/* TCP keepalive params: */
+	case OSMO_STREAM_SRV_PAR_TCP_SOCKOPT_KEEPALIVE:
+		if (!val || val_len != sizeof(uint8_t))
+			return -EINVAL;
+		on = !!*(uint8_t *)val;
+		return stream_setsockopt_tcp_keepalive(osmo_stream_srv_get_fd(conn), on);
+	case OSMO_STREAM_SRV_PAR_TCP_SOCKOPT_KEEPIDLE:
+		if (!val || val_len != sizeof(int))
+			return -EINVAL;
+		i = *(int *)val;
+		return stream_setsockopt_tcp_keepidle(osmo_stream_srv_get_fd(conn), i);
+	case OSMO_STREAM_SRV_PAR_TCP_SOCKOPT_KEEPINTVL:
+		if (!val || val_len != sizeof(int))
+			return -EINVAL;
+		i = *(int *)val;
+		return stream_setsockopt_tcp_keepintvl(osmo_stream_srv_get_fd(conn), i);
+	case OSMO_STREAM_SRV_PAR_TCP_SOCKOPT_KEEPCNT:
+		if (!val || val_len != sizeof(int))
+			return -EINVAL;
+		i = *(int *)val;
+		return stream_setsockopt_tcp_keepcnt(osmo_stream_srv_get_fd(conn), i);
+	case OSMO_STREAM_SRV_PAR_TCP_SOCKOPT_USER_TIMEOUT:
+		if (!val || val_len != sizeof(unsigned int))
+			return -EINVAL;
+		u = *(unsigned int *)val;
+		return stream_setsockopt_tcp_user_timeout(osmo_stream_srv_get_fd(conn), u);
+		break;
+	default:
+		return -ENOENT;
+	};
+}
+
 /*! @} */
